/*
 * tcp_client.cpp
 *
 *  Created on: 03.01.2021
 *      Author: doralitze
 */

#include "net/tcp_client.hpp"

#include <ev++.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <utility>
#include <list>

#include "net/netio_exception.hpp"
#include "net/socketaddress.hpp"
#include "net/sock_address_factory.hpp"

namespace rmrf::net {

tcp_client::tcp_client(
    auto_fd&& socket_fd_,
    const socketaddr& own_address_,
    const socketaddr& peer_address_,
    const std::shared_ptr<::ev::loop_ref>& loop_
) :
    connection_client{std::forward<auto_fd>(socket_fd_), peer_address_, loop_}
{
    // TODO query own socket name (in the future this will be the only construcor, as we are generated by the client_factory, io is initialized in connection_client and we query the appropiate own endpoint here)
    // TODO log created client
    this->own_address = own_address_;
}

tcp_client::~tcp_client() {}

ssize_t tcp_client::push_write_queue(::ev::io& w, iorecord& buffer) {
    return write(w.fd, buffer.ptr(), buffer.size());
}

inline uint16_t tcp_client::get_port() {
    const auto optional_port = this->peer_address.get_port_if_applicable();
    if (!optional_port)
        throw netio_exception("The associated peer (if any) does not seam to support the concept of a port.");
    return optional_port.value();
}

inline std::string get_recv_error() {
    switch (errno) {
    case EBADF:
        return "The argument sockfd is an invalid file descriptor.";
    case ECONNREFUSED:
        return "A remote host refused to allow the network connection.";
    case EFAULT:
        return "The receive buffer pointer(s) point outside the process's address space.";
    case EINTR:
        return "The receive was interrupted by delivery of a signal before any data was available.";
    case EINVAL:
        return "Invalid argument passed.";
    case ENOMEM:
        return "Could not allocate memory for recvmsg().";
    case ENOTCONN:
        return "The  socket  is  associated  with a connection-oriented protocol and has not been connected.";
    case ENOTSOCK:
        return "The file descriptor sockfd does not refer to a socket.";
    case 104:
        return "Connection Reset by peer (Linux bug)";
    default:
        return "Error Code not captured: " + std::to_string(errno);
    }
}

void tcp_client::read_from_socket(::ev::io& w) {
    // notify incomming_data_cb
    uint8_t buffer[1024];

    ssize_t n_read_bytes = recv(w.fd, buffer, sizeof(buffer), 0);

    if (n_read_bytes < 0) {
        if (errno == ECONNRESET) {
            this->stop_server();
            return;
        } else if (errno != EAGAIN) {
            throw netio_exception("Failed to read from network socket in TCP client. " + get_recv_error());
        }
    }

    if (n_read_bytes == 0) {
        // TODO find a way to properly announce the closed connection
        this->stop_server();
        return;
    } else {
        this->in_data_cb(iorecord(buffer, n_read_bytes));
    }
}

}
